const OVERFLOW_REGEXP = /(auto|scroll)/;
function getScrollTop(el) {
    return isDocumentElement(el) ? window.pageYOffset : el.scrollTop;
}
function scrollTo(el, top) {
    isDocumentElement(el) ? window.scrollTo(0, top) : (el.scrollTop = top);
}
function isDocumentElement(el) {
    return (el === document.documentElement || el === document.body || el === window);
}
function getScrollParent(el) {
    let style = getComputedStyle(el);
    const excludeStaticParent = (style.position === 'absolute');
    if (style.position === 'fixed') {
        return document.documentElement;
    }
    for (let parent = el; (parent = parent ? parent.parentElement : null);) {
        style = getComputedStyle(parent);
        if (!(excludeStaticParent && style.position === 'static') &&
            OVERFLOW_REGEXP.test(`${style.overflow}${style.overflowY}${style.overflowX}`)) {
            return parent;
        }
    }
    return document.documentElement;
}
function smoothScrollTo(element, to, duration = 300, callback) {
    let currentTime = 0;
    const start = getScrollTop(element);
    const change = (to - start);
    const easeOutCubic = (t) => change * ((t = t / duration - 1) * t * t + 1) + start;
    function smoothScroller() {
        currentTime += 5;
        scrollTo(element, easeOutCubic(currentTime));
        if (currentTime < duration) {
            window.requestAnimationFrame(smoothScroller);
        }
        else {
            callback && callback();
        }
    }
    window.requestAnimationFrame(smoothScroller);
}
export const calculateMenuTop = (menuHeight, menuEl, controlEl) => {
    const menuHeightOrDefault = (menuHeight > 0 || !menuEl) ? menuHeight : menuEl.getBoundingClientRect().height;
    const controlHeight = controlEl ? controlEl.getBoundingClientRect().height : 0;
    const menuElStyle = menuEl && getComputedStyle(menuEl);
    const marginBottom = menuElStyle ? parseInt(menuElStyle.marginBottom || '0', 10) : 0;
    const marginTop = menuElStyle ? parseInt(menuElStyle.marginTop || '0', 10) : 0;
    return `calc(${-Math.abs(menuHeightOrDefault + controlHeight)}px + ${marginBottom + marginTop}px)`;
};
export function menuFitsBelowControl(menuEl) {
    if (!menuEl)
        return true;
    const menuRect = menuEl.getBoundingClientRect();
    const scrollParent = getScrollParent(menuEl);
    const scrollTop = getScrollTop(scrollParent);
    const scrollSpaceBelow = (scrollParent.getBoundingClientRect().height - scrollTop - menuRect.top);
    return (scrollSpaceBelow >= menuRect.height);
}
export function scrollMenuIntoViewOnOpen(menuEl, menuScrollDuration, scrollMenuIntoView, handleOnMenuOpen) {
    if (!menuEl) {
        handleOnMenuOpen();
        return;
    }
    const viewInner = window.innerHeight;
    const menuRect = menuEl.getBoundingClientRect();
    const viewSpaceBelow = viewInner - menuRect.top;
    if (viewSpaceBelow >= menuRect.height) {
        handleOnMenuOpen();
        return;
    }
    const scrollParent = getScrollParent(menuEl);
    const scrollTop = getScrollTop(scrollParent);
    const scrollSpaceBelow = (scrollParent.getBoundingClientRect().height - scrollTop - menuRect.top);
    const notEnoughSpaceBelow = scrollSpaceBelow < menuRect.height;
    if (notEnoughSpaceBelow || !scrollMenuIntoView) {
        const condensedMenuHeight = notEnoughSpaceBelow ? scrollSpaceBelow : undefined;
        handleOnMenuOpen(condensedMenuHeight);
        return;
    }
    const marginBottom = parseInt(getComputedStyle(menuEl).marginBottom || '0', 10);
    const scrollDown = (menuRect.bottom - viewInner + scrollTop + marginBottom);
    smoothScrollTo(scrollParent, scrollDown, menuScrollDuration, handleOnMenuOpen);
}
